from .factory import CamelotV3Factory, GetPoolRequest
from .pool import CamelotV3Pool, GlobalState

__all__ = [
    "CamelotV3Factory",
    "GetPoolRequest",
    "CamelotV3Pool",
    "GlobalState",
]



from typing import Annotated  # noqa: D100

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, primitives
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel

V3_FACTORY_ADDRESS = HexAddress(HexStr("0x1F98431c8aD98523631AE4a59f267346ea31F984"))


class GetPoolRequest(BaseModel):
    token_a: Annotated[HexAddress, Name("tokenA")]
    token_b: Annotated[HexAddress, Name("tokenB")]
    fee: primitives.uint24


class UniswapV3Factory(ProtocolBase):
    get_pool: Annotated[ContractFunc[GetPoolRequest, HexAddress], Name("getPool")] = (
        METHOD
    )



from typing import Annotated

from eth_rpc.event import Event
from eth_rpc.types import Indexed, Name, primitives
from eth_typing import HexAddress
from pydantic import BaseModel


class AddIndexType(BaseModel):
    index: Annotated[HexAddress, Indexed]
    verified: primitives.bool


class RemoveIndexType(BaseModel):
    index: Annotated[HexAddress, Indexed]


class SetVerifiedType(BaseModel):
    index: Annotated[HexAddress, Indexed]
    verified: primitives.bool


# from ownable
class OwnershipTransferredType(BaseModel):
    previous_owner: Annotated[HexAddress, Indexed, Name("previousOwner")]
    new_owner: Annotated[HexAddress, Indexed, Name("newOwner")]


AddIndexEvent = Event[AddIndexType](name="AddIndex")
RemoveIndexEvent = Event[RemoveIndexType](name="RemoveIndex")
SetVerifiedEvent = Event[SetVerifiedType](name="SetVerified")
OwnershipTransferredEvent = Event[OwnershipTransferredType](name="OwnershipTransferred")



from .account_manager import AccountManager
from .blocks import Blocks, NonceArgs
from .command_tracker import CommandTracker
from .helpers import TokenRegistry, TokenStruct
from .mempool import Mempool
from .oracle import AttestationOracle
from .registry import ProtocolRegistry
from .sources import CommandSource, XSource

__all__ = [
    "AccountManager",
    "AttestationOracle",
    "Blocks",
    "CommandSource",
    "CommandTracker",
    "Mempool",
    "NonceArgs",
    "ProtocolRegistry",
    "TokenRegistry",
    "TokenStruct",
    "XSource",
]



from typing import Annotated, Any, ClassVar

from eth_abi import encode
from eth_rpc import ContractFunc
from eth_rpc.types import METHOD, Name, primitives

from ..utils import BaseSource, Command, SimulacrumSubmission, Tweet


class XSource(BaseSource):
    source_name: ClassVar[str] = "X"
    input_type: ClassVar[type] = Tweet

    lookup_address: Annotated[
        ContractFunc[
            tuple[primitives.bytes32, primitives.uint256],
            primitives.address,
        ],
        Name("lookupAddress"),
    ] = METHOD
    convert: ContractFunc[
        primitives.bytes,
        Command,
    ] = METHOD
    load_command: Annotated[
        ContractFunc[
            primitives.bytes32,
            Command,
        ],
        Name("loadCommand"),
    ] = METHOD

    def submit_tweet(
        self,
        tweet: Tweet,
        signatures: list[bytes],
        require_success: bool = True,
    ) -> ContractFunc[
        SimulacrumSubmission,
        bool,
    ]:
        return self.submit(
            SimulacrumSubmission(
                source_data=tweet.to_bytes(),
                verification_data=encode(["bytes[]"], [signatures]),
                require_success=require_success,
            )
        )

    @classmethod
    def validate_input(cls, input: Any):
        if not isinstance(input, Tweet):
            return False
        text: str = input.full_text
        first_line = text.split("\n")[0]

        modifiers = first_line.split(" ")
        if len(modifiers) < 2:
            return False
        if modifiers[0] != "#simulacrum":
            return False
        return True



from typing import Annotated

from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name

from .types import (
    AddLiquidityETHRequest,
    AddLiquidityRequest,
    AddLiquidityResponse,
    EthSwapRequest,
    RemoveLiquidityETHRequest,
    RemoveLiquidityETHResponse,
    RemoveLiquidityRequest,
    RemoveLiquidityResponse,
    TokenSwapRequest,
)

ROUTER_ADDRESS = "0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D"


class UniswapV2Router(ProtocolBase):
    add_liquidity: Annotated[
        ContractFunc[
            AddLiquidityRequest,
            AddLiquidityResponse,
        ],
        Name("addLiquidity"),
    ] = METHOD
    add_liquidity_eth: Annotated[
        ContractFunc[
            AddLiquidityETHRequest,
            AddLiquidityResponse,
        ],
        Name("addLiquidityETH"),
    ] = METHOD

    remove_liquidity: Annotated[
        ContractFunc[
            RemoveLiquidityRequest,
            RemoveLiquidityResponse,
        ],
        Name("removeLiquidity"),
    ]
    remove_liquidity_eth: Annotated[
        ContractFunc[
            RemoveLiquidityETHRequest,
            RemoveLiquidityETHResponse,
        ],
        Name("removeLiquidityETH"),
    ]

    swap_exact_tokens_for_tokens_supporting_fee_on_transfer_tokens: Annotated[
        ContractFunc[
            TokenSwapRequest,
            None,
        ],
        Name("swapExactTokensForTokensSupportingFeeOnTransferTokens"),
    ]

    swap_exact_eth_for_tokens_supporting_fee_on_transfer_tokens: Annotated[
        EthSwapRequest,
        Name("swapExactETHForTokensSupportingFeeOnTransferTokens"),
    ]

    # TODO: finish adding methods



from typing import Annotated, Any

from eth_abi import encode
from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.types import METHOD, Name, NoArgs, primitives

from ...models import Command, SimulacrumSubmission, Tweet


class XSource(ProtocolBase):
    convert: ContractFunc[bytes, Command] = METHOD

    default_gas: Annotated[
        ContractFunc[NoArgs, primitives.uint256],
        Name("defaultGas"),
    ] = METHOD

    load_command: Annotated[
        ContractFunc[primitives.bytes32, Command],
        Name("loadCommand"),
    ] = METHOD

    lookup_address_with_index: Annotated[
        ContractFunc[tuple[primitives.bytes32, primitives.uint256], primitives.address],
        Name("lookupAddress"),
    ] = METHOD

    lookup_address: Annotated[
        ContractFunc[primitives.bytes32, primitives.address],
        Name("lookupAddress"),
    ] = METHOD

    max_gas: Annotated[
        ContractFunc[NoArgs, primitives.uint256],
        Name("maxGas"),
    ] = METHOD

    oracle: ContractFunc[NoArgs, primitives.address] = METHOD

    submit: ContractFunc[SimulacrumSubmission, Annotated[bool, "success"]] = METHOD

    submit_many: Annotated[
        ContractFunc[tuple[list[bytes], list[bytes], bool], None],
        Name("submitMany"),
    ] = METHOD

    update_oracle: Annotated[
        ContractFunc[primitives.address, None],
        Name("updateOracle"),
    ] = METHOD

    verify: ContractFunc[tuple[bytes, bytes], bool] = METHOD

    def submit_tweet(
        self,
        tweet: Tweet,
        signatures: list[bytes],
        require_success: bool = True,
    ) -> ContractFunc[
        SimulacrumSubmission,
        bool,
    ]:
        return self.submit(
            SimulacrumSubmission(
                source_data=tweet.to_bytes(),
                verification_data=encode(["bytes[]"], [signatures]),
                require_success=require_success,
            )
        )

    @classmethod
    def validate_input(cls, input: Any):
        if not isinstance(input, Tweet):
            return False
        text: str = input.full_text
        first_line = text.split("\n")[0]

        modifiers = first_line.split(" ")
        if len(modifiers) < 2:
            return False
        if modifiers[0] != "#simulacrum":
            return False
        return True



"""
Orbiter has a unique technique for bridging.  An account can transfer
funds directly to the Orbiter address for the asset, and the trailing
4 digits of the transaction is used to identify the target network.

For example, transferring 100009001 ETH to 0x80c67432656d59144ceff962e8faf8926599bcf8
on any supported chain will bridge the funds (minus a fee) to Mainnet.
"""


class OrbiterRelays:
    USDC_ADDRESS = "0x41d3D33156aE7c62c094AAe2995003aE63f587B3"
    ETH_ADDRESS = "0x80c67432656d59144ceff962e8faf8926599bcf8"
    USDT_ADDRESS = "0x80c67432656d59144ceff962e8faf8926599bcf8"


class OrbiterChainIds:
    MAINNET: int = 9001
    ARBITRUM: int = 9002
    BASE: int = 9021
    LINEA: int = 9023



from typing import Annotated

from eth_rpc.event import Event
from eth_rpc.types import Indexed, primitives
from eth_typing import HexAddress
from pydantic import BaseModel


class BondType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    token: Annotated[HexAddress, Indexed]
    amount_tokens_bonded: primitives.uint256
    amount_tokens_minted: primitives.uint256


class DebondType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    amount_debonded: primitives.uint256


class AddLiquidityType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    amount_tokens: primitives.uint256
    amount_DAI: primitives.uint256


class RemoveLiquidityType(BaseModel):
    wallet: Annotated[HexAddress, Indexed]
    amount_liquidity: primitives.uint256


class FlashLoanType(BaseModel):
    executor: Annotated[HexAddress, Indexed]
    recipient: Annotated[HexAddress, Indexed]
    token: HexAddress
    amount: primitives.uint256


class CreateType(BaseModel):
    new_idx: Annotated[HexAddress, Indexed]
    wallet: Annotated[HexAddress, Indexed]


CreateEvent = Event[CreateType](name="Create")
FlashLoanEvent = Event[FlashLoanType](name="FlashLoan")
AddLiquidityEvent = Event[AddLiquidityType](name="AddLiquidity")
RemoveLiquidityEvent = Event[RemoveLiquidityType](name="RemoveLiquidity")
BondEvent = Event[BondType](name="Bond")
DebondEvent = Event[DebondType](name="Debond")



import math

from ..constants import Q96

TICK_SPACING = {
    10_000: 200,
    3_000: 60,
    500: 10,
    100: 1,
}


def tick_to_price(tick, token0_decimals=18, token1_decimals=18) -> float:
    return 1.0001**tick * (10 ** (token0_decimals - token1_decimals))


def get_tick_at_sqrt_price(sqrt_price_x96) -> int:
    return math.floor(math.log((sqrt_price_x96 / Q96) ** 2) / math.log(1.0001))


def get_closest_low_tick(tick: int, tick_spacing: int = 200) -> int:
    # Ensure tickSpacing is not zero
    if tick_spacing == 0:
        raise ValueError("tickSpacing cannot be zero")

    remainder = tick % tick_spacing
    if remainder < 0:
        return (tick // tick_spacing) * tick_spacing - tick_spacing
    else:
        return (tick // tick_spacing) * tick_spacing


def get_closest_high_tick(tick: int, tick_spacing: int = 200) -> int:
    # Ensure tickSpacing is not zero
    if tick_spacing == 0:
        raise ValueError("tickSpacing cannot be zero")

    closest_low_tick = get_closest_low_tick(tick, tick_spacing)

    if tick % tick_spacing == 0:
        return closest_low_tick  # Tick is exactly on a spacing boundary
    else:
        return closest_low_tick + tick_spacing



from collections.abc import Awaitable
from typing import Annotated, Any, Generic, Literal, TypeVar, overload

from eth_abi.exceptions import InsufficientDataBytes
from eth_rpc import ContractFunc, ProtocolBase
from eth_rpc.networks import Network
from eth_rpc.types import (
    BLOCK_STRINGS,
    METHOD,
    MaybeAwaitable,
    Name,
    Struct,
    primitives,
)
from eth_rpc.utils import handle_maybe_awaitable, run
from eth_typing import HexAddress, HexStr
from pydantic import BaseModel, Field

T = TypeVar("T", bound=tuple[Any])
U = TypeVar("U")

MULTICALL3_ADDRESS = HexAddress(HexStr("0xcA11bde05977b3631167028862bE2a173976CA11"))


class Result(Struct):
    success: bool
    return_data: bytes


class TryMulticallRequest(BaseModel):
    require_success: Annotated[bool, Name("requireSuccess")]
    calls: list[tuple[primitives.address, primitives.bytes]]


class MulticallRequest(BaseModel):
    calls: list[tuple[primitives.address, primitives.bytes]]


class TryAggregateResponse(BaseModel):
    results: list[Result]


class TryMulticallResponse(BaseModel):
    block_number: Annotated[primitives.uint256, Name("blockNumber")]
    block_hash: Annotated[primitives.bytes32, Name("blockHash")]
    return_data: Annotated[
        list[tuple[primitives.bool, primitives.bytes]], Name("returnData")
    ]


class TryResult(BaseModel, Generic[T]):
    success: bool
    result: T | Any


class Multicall(ProtocolBase):
    address: HexAddress = Field(default=MULTICALL3_ADDRESS)

    block_and_aggregate: Annotated[
        ContractFunc[
            MulticallRequest,
            TryMulticallResponse,
        ],
        Name("blockAndAggregate"),
    ] = METHOD
    try_aggregate: Annotated[
        ContractFunc[
            TryMulticallRequest,
            TryAggregateResponse,
        ],
        Name("tryAggregate"),
    ] = METHOD
    try_block_and_aggregate: Annotated[
        ContractFunc[
            TryMulticallRequest,
            TryMulticallResponse,
        ],
        Name("tryBlockAndAggregate"),
    ] = METHOD

    async def _execute(
        self,
        *calls: ContractFunc,
        sync: bool = False,
        block_number: int | BLOCK_STRINGS = "latest",
    ) -> list[Any]:
        results = await handle_maybe_awaitable(
            self.try_execute(
                *calls, require_success=True, block_number=block_number, sync=sync
            )
        )
        return [r.result for r in results]

    @overload
    def execute(
        self,
        *calls: ContractFunc,
        sync: Literal[False],
        block_number: int | BLOCK_STRINGS = ...,
    ) -> list[Any]: ...

    @overload
    def execute(
        self,
        *calls: ContractFunc,
        sync: Literal[True],
        block_number: int | BLOCK_STRINGS = ...,
    ) -> Awaitable[list[Any]]: ...

    @overload
    def execute(
        self,
        *calls: ContractFunc,
        block_number: int | BLOCK_STRINGS = ...,
    ) -> Awaitable[list[Any]]: ...

    def execute(
        self,
        *calls: ContractFunc,
        block_number: int | BLOCK_STRINGS = "latest",
        sync: bool = False,
    ) -> MaybeAwaitable[list[Any]]:
        return run(self._execute, *calls, block_number=block_number, sync=sync)

    async def _try_execute(
        self,
        *calls: ContractFunc,
        require_success: bool = False,
        sync: bool = False,
        block_number: int | BLOCK_STRINGS = "latest",
    ) -> list[TryResult]:
        self.try_aggregate._network = self._network
        call = await self.try_aggregate(
            TryMulticallRequest(
                require_success=require_success,
                calls=[(c.address, c.encode()) for c in calls],
            )
        ).call(sync=sync, block_number=block_number)

        return_data = call.decode()
        response: list[TryResult] = []

        for result, func in zip(return_data.results, calls):
            if result.success:
                try:
                    response.append(
                        TryResult(success=True, result=func.decode(result.return_data))
                    )
                except (OverflowError, InsufficientDataBytes):
                    response.append(TryResult(success=False, result=None))
            else:
                response.append(TryResult(success=False, result=None))
        return response

    @overload
    def try_execute(
        self,
        *calls: ContractFunc,
        sync: Literal[True],
        require_success: bool = ...,
        block_number: int | BLOCK_STRINGS = ...,
    ) -> list[TryResult]: ...

    @overload
    def try_execute(
        self,
        *calls: ContractFunc,
        require_success: bool = ...,
        block_number: int | BLOCK_STRINGS = ...,
    ) -> Awaitable[list[TryResult]]: ...

    @overload
    def try_execute(
        self,
        *calls: ContractFunc,
        require_success: bool = ...,
        sync: bool = ...,
        block_number: int | BLOCK_STRINGS = ...,
    ) -> MaybeAwaitable[list[TryResult]]: ...

    def try_execute(
        self,
        *calls: ContractFunc,
        require_success: bool = False,
        sync: bool = False,
        block_number: int | BLOCK_STRINGS = "latest",
    ) -> MaybeAwaitable[list[TryResult]]:
        return run(
            self._try_execute,
            *calls,
            require_success=require_success,
            sync=sync,
            block_number=block_number,
        )


multicall = Multicall(address=MULTICALL3_ADDRESS)


def make_multicall(network: type[Network]):
    return Multicall[network](address=MULTICALL3_ADDRESS)
